'''
Reduces (i.e. detunes) 405 laser throughput by changing AOTF setting.

License: Public Domain
Author: Pariksheet Nanda <p.nanda@andor.com> 2013-11-11
'''

from __future__ import division # for linspace()
import logging
import re
from serial import Serial
from time import sleep

logging.basicConfig(level=logging.ERROR)


class RS232(Serial):
    '''Convenience serial wrapper.'''
    WRITE_DELAY_SEC = 0.1
    READ_DELAY_SEC = 0.1

    def __init__( self, *p, **kw):
        super().__init__(*p, **kw)

    def receive(self):
        '''Returns contents of receive buffer as byte string.'''
        received = ''
        while self.inWaiting() > 0:
            logging.debug('Reading {0} bytes...'.format(self.inWaiting()))
            received += str(self.read(self.inWaiting()))
            sleep(self.READ_DELAY_SEC)
        return received

    def send(self, value):
        '''Send string value and return byte string.'''
        self.write(value)
        sleep(self.WRITE_DELAY_SEC)
        received = self.receive()
        logging.debug('Raw received bytes from sending' +\
                          '"{0}": {1}'.format(value, received))
        return received


class AAAOTF(RS232):
    '''AA Opto Electronics AOTF controller.'''
    _power = 0

    def __init__(self, port):
        '''Connect to AOTF controller.'''
        super().__init__(port=port, baudrate=19200, timeout=1)

    def _allowed_powers(self):
        return [12.0, 12.2, 12.3, 12.5, 12.6, 12.8, 13.0, 13.2, 13.3, 13.5,
                13.6, 13.8, 14.0, 14.1, 14.3, 14.4, 14.6, 14.8, 15.0, 15.1,
                15.2, 15.4, 15.6, 15.8, 15.9, 16.1, 16.3, 16.4, 16.6, 16.7,
                16.9, 17.1, 17.2, 17.4, 17.5, 17.7, 17.9, 18.1, 18.2, 18.4,
                18.5, 18.7, 18.9, 19.0, 19.2, 19.3, 19.5, 19.7, 19.9, 20.0,
                20.2, 20.3, 20.5, 20.7, 20.8, 21.0, 21.2, 21.3, 21.5, 21.7,
                21.8, 22.0, 22.1, 22.3]

    def _power_index(self, value):
        '''Converts dB power to 0-63 index for the controller.'''
        value = float(value)
        powers = self._allowed_powers()
        logging.info('Searching for power {0} dB in allowed list...'.format(value))
        if powers.count(value):
            logging.debug('...exact match found.')
            return powers.index(value)
        else:
            # Find the closest value in the list.
            powers2 = powers + [value]
            powers2.sort()
            i = powers2.index(value)
            left_power = powers2[i-1]
            right_power = powers2[i+1]
            left_diff = round(abs(value - left_power), 1)
            right_diff = round(abs(value - right_power), 1)
            logging.debug('-{0}/+{1}'.format(left_diff, right_diff))
            if left_diff == right_diff:
                if (value * 10) % 2 == 0:
                    logging.debug('...closest lower match is {0} dB.'.format(left_power))
                    return powers.index(left_power)
                else:
                    logging.debug('...closest upper match is {0} dB.'.format(right_power))
                    return powers.index(right_power)
            elif left_diff < right_diff:
                logging.debug('...closest match is {0} dB.'.format(left_power))
                return powers.index(left_power)
            else:
                logging.debug('...closest match is {0} dB.'.format(right_power))
                return powers.index(right_power)

    @property
    def power(self):
        '''Returns channel power in dB.'''
        received = self.send('S'.encode('latin1'))
        # received is '\rl1 F=152.103 P=19.4...'
        match = re.search('(?<=P[=])\d{2}[.]\d', received)
        self._power = float(match.group(0))
        return self._power

    @power.setter
    def power(self, value):
        '''Sets channel power in dB.'''
        self.send('P'.encode('latin1'))
        self.send(str(self._power_index(value)).encode('latin1'))
        self.send('\r'.encode('latin1'))

    def power_from_index(self, value):
        '''Sets channel power using index.'''
        self.send('P'.encode('latin1'))
        self.send(str(value).encode('latin1'))
        self.send('\r'.encode('latin1'))
        return self.power

if __name__ == '__main__':
    ser = AAAOTF('COM1')
    try:
        print('Old power: {0}'.format(ser.power))
        set = 18.9
        print('Set power: {0}'.format(set))
        ser.power = set
        print('New power: {0}'.format(ser.power))

        # print('Generating list of possible powers')
        # for i in range(64):
        #     power = ser.power_from_index(i)
        #     print(power, end=', ', flush=True)
    finally:
        ser.close()
